/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/>
 * ****************************************************************************/


/*
 * 
 * Dodwan websocket YJS provider.
 * Class DodwanProvider allows the instanciation of providers for YJS that uses DoDWAN to propagate updates.
 * A provider is linked to a unique YJS doc (associated with the doc by a unique doc name).
 * When a provider connects (call to connect()), a DoDWAN NAPI session is opened (named after the doc name), 
 * using a websocket connection to the DoDWAN daemon.
 * The provider implement two synchronization protocol, one (delta-state based) with other replicas, 
 * i.e, with other providers, and one with with relays.
 * The provider listens to events related to 
 *  - the session and the connection.
 *  - the peers (replica peers and relay peers, relay peers having a name starting with "R_")
 *  - update events related to the YJS doc it is attached to: it broadcasts a local update
 *    to the current replica neighbors when this update occurs.
 * When encountering a replica neighbor, the provider exchanges its delta-state with this neighbor (the replica with lower name starts the exchange). 
 * When encountering a relay neighbor, the provider exchanges, if it is useful, its state with the ones hold by the relay (the replica starts the exchange)
 * When connecting, the provider exchanges its delta-state with each of its replica neighbors and it state with each of its relay neighbors. 
 * As a local update is only broadcast to other replicas, the synchronization with a relay is also performed periodically, as long as the relay remains a neighbor. 
 */


import * as napicli from "@casa/napi-ws-client"
import * as Y from "yjs"
import {VersionVector, hash, isRelay} from "@casa/relay-util" 


const RELAY_EXCG_PERIOD = 17000; // Period for synchro with connected relays (ms)


console.log("Start module yjs-ws-provider");

 function log(src, msg) {
       napicli.util.serviceLog(src, 'yjs_provider', msg);
 }



/*
 * Class DodwanProvider
 * @param {string} serverURL -  URL of the form [ws://]host[:port]  (no path should be given, parsing not robust)
 * @param {string} ydocname - The Yjs document name
 * @param {Y.Doc} ydoc - The Yjs document
 * @param {Object} ops - Options
 * @param {boolean} ops.connect - set to false if you want to connect manually (@see connect())
 * @param {boolean} ops.strvv - initial value of the version vector (string form)
 *    
 * NOTA: The signature of this constructor mimics the one of the standard YJS ws-provider   
 */
export class DodwanProvider  {
    
    constructor(serverURL, ydocname, ydoc, opts={connect:true, strvv:"{}"}) {
        //log(this.myPeerName+"("+this.ydocname+")", null, "Dodwanprovider created with opts="+JSON.stringify(opts));
        this.shouldConnect = opts.connect;
        
        this.sessionListener;
        this.peerListener;
        this.myPeerName;
        
        // Map peer => timeoutId, used for periodic synchro with relays (which may be stable neighbors)
        this.pendingSynchroPeers = new Map();  

        // Used to remember data fom previous messages in the synchro protocol with replicas
        this.responseCallbacks = new Map();  // <peer String, callback Function()>
        this.stateVectors = new Map();       // <peer String, stvector Uint8Array>
               
        this.ydoc = ydoc;
        this.ydocname = ydocname;
        this.ydocid = "id" + ydoc.clientID;  
     
        
        let hostport;
        let slashpos = serverURL.indexOf('//');
        if (slashpos===-1) {
           hostport = serverURL;
        }
        else {
            hostport=serverURL.substring(slashpos+2);
        }
        let colonpos=hostport.lastIndexOf(':');
        if (colonpos===-1) {
            this.wshost=hostport;
            this.wsport=8025;
        }
        else {
            this.wshost = hostport.substring(0,colonpos);
            this.wsport = hostport.substring(colonpos+1);
        }
        
        // Version vector (actually the version vector of local updates) that serves for the synchro with relays
        this.versionVector = new VersionVector(opts.strvv);

        log(this.myPeerName+"("+this.ydocname+")", "creating session for Ydoc docname="+this.ydocname+"  wsport="+this.wsport+ " connect="+this.shouldConnect);
        
        this.configuration = new napicli.napiConf.Configuration();
        this.session = new napicli.napiSession.Session(this.configuration);
        this.peerService = new napicli.napiPeer.PeerService(this.session);
       
        // Connect if not specifically asked not to do so
        if (this.shouldConnect) {
            //log(this.myPeerName+"("+this.ydocname+")", "connecting");
            this.connect();
        }
            
    }
    
    
    /* ----------------------------------------------------------------------
    * Handler for session started
    */
    onSessionStarted = () => {
        //log(this.myPeerName+"("+this.ydocname+")", "Session started"); 
    
        // Prepare to receive DoDWAN D2D messages from peers
        //log(this.myPeerName+"("+this.ydocname+")", "registering handler for ReceiveD2dNotification");
        this.session.setOnReceiveD2dNotification(
           (pdu) => { 
            let edp = pdu.desc._endpoint;
            if (edp !== undefined) {
                //log(this.myPeerName+"("+this.ydocname+")", "received D2D notif for endpoint doc" +edp);
                this.messageDescProcessor(pdu.desc); 
            }
            else {
                log(this.myPeerName+"("+this.ydocname+")", "ignoring incoming PDU "+JSON.stringify(pdu.desc));
            }
        });
    
        // Prepare to react on doc changes
        this.startReactOnDocChanges(); 
    
        // Get my name and start exchanges with neighbors
        this.peerService.getMyPeerName(
           (peer) => {
               this.myPeerName = peer;
                log(this.myPeerName+"("+this.ydocname+")", "I am peer "+this.myPeerName);
                this.peerService.getPeers(
                    (peers) => {
                         this.delayedExchangesWithReplicas(peers);
                         this.delayedExchangesWithRelays(peers);
                    }); 
            }
        );
  
    }
    
    
    /**
    * Initiates a periodic exchange with neighboring relays.
    * Each conversation is sligthly shifted in time
    * @param {[String] peers
    * @returns {undefined}
    */
    delayedExchangesWithRelays(peers)  {
        const SHIFT_DELAY = 55; 
        let delay = 0;
        for (const p of peers) {
            log("Found peer "+p);
            if (isRelay(p)) {
                //log("Programming synchro with relay "+p+" in "+delay + " ms");
                setTimeout( () => this.synchroWithRelay(p), delay );
            }
            delay += SHIFT_DELAY ;
        }
    }
 
    
     /**
    * Initiates exchanges with neighboring replicas.
    * Each conversation is sligthly shifted in time
    * @param {[String] peers
    * @returns {undefined}
    */
    delayedExchangesWithReplicas(peers)  {
        const SHIFT_DELAY = 50; 
        let delay = 0;
        for (const p of peers) {
            log("Found peer "+p);
            if (isNotRelay(p)) {
                //log("Programming synchro with replica "+p+" in "+delay + " ms");
                setTimeout( () => this.exchangeDeltaWithPeer(p), delay );
                delay += SHIFT_DELAY ;
            }
     
            
        }
    }
 

    /* ----------------------------------------------------------------------
    * Handler for session stopped
    */
    onSessionStopped = () => {
        log(this.myPeerName+"("+this.ydocname+")", "session stopped"); 
        this.stopSynchroWithRelays();
        this.stopReactOnDocChanges();  
        this.stateVectors.clear();
        this.responseCallbacks.clear();
    }


    /* ----------------------------------------------------------------------
    * Other status handlers
    */
    onSessionStarting() {log(this.myPeerName+"("+this.ydocname+")", "session starting");}
    onConnectionError() {log(this.myPeerName+"("+this.ydocname+")", "connection error");}
    onConnectionOpened() {log(this.myPeerName+"("+this.ydocname+")", "connection opened");}
    onConnectionClosed() {log(this.myPeerName+"("+this.ydocname+")", "connection closed");} 
 
    
    
    /* ----------------------------------------------------------------------
    * Initiates the NAPI session 
    */
    initNapiSession(wshost, wsport, ydocname) {
        // Register the configuration of the Websocket server
        log(this.myPeerName+"("+this.ydocname+")", "configure NAPI for server host="+wshost+ "  port="+wsport);
        this.configuration.init({"server_host":wshost, "server_port":wsport, "name":ydocname});
        
        // Register peers event handlers
        this.peerListener = new napicli.napiPeer.PeerListener(
              this.addPeer, this.removePeer, this.clearPeers
        );
        this.peerService.addListener(this.peerListener);

        // Register session event handlers
        this.sessionListener = new napicli.napiSession.SessionListener(
                        this.onConnectionOpened, this.onConnectionClosed, this.onConnectionError,
                        this.onSessionStarting, this.onSessionStarted, this.onSessionStopped
        );
        this.session.addListener(this.sessionListener);

        // Start the Napi session
        this.session.start();
        log(this.myPeerName+"("+this.ydocname+")", "waiting for NAPI session to start");
    }


  
    /* ----------------------------------------------------------------------
    * Closes the NAPI session
    */
    endNapiSession() {
        // Unregister peers event handlers
        this.peerService.removeListener(this.peerListener);

        // Unregister session event handlers
        this.session.removeListener(this.sessionListener);

        // Stop the Napi session
        this.session.bye();
        log(this.myPeerName+"("+this.ydocname+")", "NAPI session ended");
    }
  
    
    /*
     * Compatibility with ws-provider functions 
     */
    connect() {
        log(this.myPeerName+"("+this.ydocname+")", "connecting to DoDWAN ws server");
        this.shouldConnect = false;
        this.initNapiSession(this.wshost, this.wsport, this.ydocname);
    }
    
    
    disconnect() {
        log(this.myPeerName+"("+this.ydocname+")", "disconnecting from DoDWAN ws server");
        this.endNapiSession();
        this.shouldConnect = true;
    }
    
    
    /* ----------------------------------------------------------------------
    * Prepares to react to YJS doc changes
    * An 'update' event is raised when the application locally updates the YJS doc 
    * but also when the Yjs provider applies an update that has just arrived from a peer. 
    * The update should be broadcast to peers only in the first case (no transitivity).
    * The version vector is to be incremented only when the update is local 
    * (besides, after the reception of repdiff in messageDescProcessor(), the incoming version vector 
    * is merged with the local one).
    */
    startReactOnDocChanges() {
        //log(this.myPeerName+"("+this.ydocname+")", "start reacting on changes on Ydoc "+this.ydocname);
        this.ydoc.on('update', (update, origin) => {
            
            // Don't broadcast to peers the updates that are received from peers
            if (origin !== this) {
                // this is a local update
                //log(this.myPeerName+"("+this.ydocname+")", "Inc version vector, now="+this.versionVector.toString());
                this.versionVector.inc(this.myPeerName); 
                this.broadcastUpdateToPeers(update);
            }
        });
        this.ydoc.on('destroy', doc => {this.destroyMyself();});        
    }
   
   
   /* ----------------------------------------------------------------------
    * Stops reacting to YJS doc changes
    */
    stopReactOnDocChanges() {
        //log(this.myPeerName+"("+this.ydocname+")", "stop reacting on changes on Ydoc "+this.ydocname);
        this.ydoc.off('update');
        this.ydoc.off('destroy');        
    }
 


    /**
     * Checks if a peer is still a neighbor peer,
     * synchronizes with it and programs a new synchro later.
     * @param {String} peer
     * @returns {undefined}
     */
    synchroWithRelay(peer) {
        if (this.peerService.getKnownPeers().indexOf(peer) !== -1 ) {
            this.exchangeStatesWithRelay(peer);
            this.pendingSynchroPeers.set(peer,      
                setTimeout( () => {this.synchroWithRelay(peer);},
                            RELAY_EXCG_PERIOD
                          )
            );
        }
    }


    /**
     * Cancels the pending synchros with relays
     * @returns {undefined}
     */
    stopSynchroWithRelays() {
        this.pendingSynchroPeers.foreach((value) => clearTimeout(value));
    }
    

   /* ----------------------------------------------------------------------
    * Broadcasts an update to all the replica peers
    * @param  {uint8Array} update
    */
    broadcastUpdateToPeers(update) {
        if (this.shouldconnect) return;
        let desc = new napicli.napiPubsub.Descriptor();
        desc["_type"] = "yjs";
        desc["_endpoint"] = this.ydocname;  
        desc["_ytype"] = "update";
        desc["_update"] = napicli.util.uint8ArrayToB64(update);
        //log(this.myPeerName+"("+this.ydocname+")", "bcast_update size="+update.length);
        this.broadcastToPeers(desc, isNotRelay); 
    }

    
   
   /* ----------------------------------------------------------------------
    * 
    * Exchanges delta state with peer. 
    * First exchange state vectors and then exchange diffs
    * The synchro exchange is identified by the peer identifier that serves as a token,
    * in order to remember data received in previous messages during the synchro.
    */
    exchangeDeltaWithPeer(peer) {
        if (this.shouldconnect) { log(this.myPeerName+"("+this.ydocname+")", "exchg_delta canceled (not connected)"); return;};
        //log(this.myPeerName+"("+this.ydocname+")", "exchange delta with peer "+ peer);    
        const vector = Y.encodeStateVector(this.ydoc);
        this.sendReqVectToPeer(peer,vector);          
    }
    
    
    sendReqVectToPeer(peer, vector) {  
        let desc = new napicli.napiPubsub.Descriptor();
        desc["_type"] = "yjs";
        desc["_endpoint"] = this.ydocname;  
        desc["_ytype"] = "reqvect";
        desc["_vector"] = napicli.util.uint8ArrayToB64(vector);  
        //log(this.myPeerName+"("+this.ydocname+")", "snd_reqvect dst="+peer+ ",size="+vector.length+ ",total="+this.ydoc.getText('quill').length);
        this.sendToPeer(peer, desc);       
    }
 
    sendRepVectToPeer(peer, vector) {  
        let desc = new napicli.napiPubsub.Descriptor();
        desc["_type"] = "yjs";
        desc["_endpoint"] = this.ydocname;  
        desc["_ytype"] = "repvect";
        desc["_vector"] = napicli.util.uint8ArrayToB64(vector);    
        //log(this.myPeerName+"("+this.ydocname+")", "snd_repvect dst=" + peer + ",size="+vector.length+ ",total="+this.ydoc.getText('quill').length);
        this.sendToPeer(peer, desc);       
    }
    
    sendReqDiffToPeer(peer, diff) {
        let desc = new napicli.napiPubsub.Descriptor();
        desc["_type"] = "yjs";
        desc["_endpoint"] = this.ydocname;  
        desc["_ytype"] = "reqdiff";
        desc["_diff"] = napicli.util.uint8ArrayToB64(diff);
        //const fullsize = Y.encodeStateAsUpdate(this.ydoc).length;
        //log(this.myPeerName+"("+this.ydocname+")", "snd_reqdiff dst="+peer+ ",size="+diff.length+ ",full="+fullsize+ ",total="+this.ydoc.getText('quill').length);
        this.sendToPeer(peer, desc);  
    }
     
    sendRepDiffToPeer(peer, diff) {
        let desc = new napicli.napiPubsub.Descriptor();
        desc["_type"] = "yjs";
        desc["_endpoint"] = this.ydocname;  
        desc["_ytype"] = "repdiff";
        desc["_diff"] = napicli.util.uint8ArrayToB64(diff);
        desc["_vv"] = this.versionVector.toString();
        //const fullsize = Y.encodeStateAsUpdate(this.ydoc).length;
        //log(this.myPeerName+"("+this.ydocname+")", "snd_repdiff dst=" + peer + ",size="+diff.length+ ",full="+fullsize+ ",total="+this.ydoc.getText('quill').length);
        this.sendToPeer(peer, desc);  
    }
    
    
   
    /*
     * Processes the arrival of a message (PDU recv_from_peer).
     * This message holds either
     *    - a Yjs doc op-based update from a replica (update)
     *    - one of the exchange messages for delta-state synchro from another replica (reqvect, repvect, reqdiff, repdiff)
     *    - a state message in the context of a synchro with a relay (strep)
     *    
     * @param {Descriptor} desc - the received message as a descriptor 
     */
     messageDescProcessor(desc) {
        log(this.myPeerName+"("+this.ydocname+")", "Message received: "+JSON.stringify(desc));
        
        /////////////////////////////////////// SYNCHRO WITH REPLICA /////////////////////////////////////////
       
        if (desc._ytype === "update") { // broadcast update (op-based)
            const upd = napicli.util.b64ToUint8Array(desc._update);
            this.processYjsUpdate(upd);
            //log(this.myPeerName+"("+this.ydocname+")", "rcv_"+desc._ytype+" sdr=" + desc._sdr + ",size="+upd.length + ",total="+this.ydoc.getText('quill').length);
        }

        else if (desc._ytype === "reqvect") {
            const vector = napicli.util.b64ToUint8Array(desc._vector);
            this.stateVectors.set(desc._sdr, vector); // Remember the received vector to later use it for computing the diff
            //log(this.myPeerName+"("+this.ydocname+")", "rcv_reqvect sdr=" + desc._sdr + ",size="+vector.length + ",total="+this.ydoc.getText('quill').length);
            this.sendRepVectToPeer(desc._sdr, Y.encodeStateVector(this.ydoc));
        }
        
        else if (desc._ytype === "repvect") {
            const vector =  napicli.util.b64ToUint8Array(desc._vector);
            //log(this.myPeerName+"("+this.ydocname+")", "rcv_repvect sdr=" + desc._sdr + ",size="+vector.length + ",total="+this.ydoc.getText('quill').length); 
            const diff = Y.encodeStateAsUpdate(this.ydoc, vector);
            this.sendReqDiffToPeer(desc._sdr, diff);       
        }
                     
        else if (desc._ytype === "reqdiff")  {
            const diff = napicli.util.b64ToUint8Array(desc._diff);
            this.sendRepDiffToPeer(desc._sdr, Y.encodeStateAsUpdate(this.ydoc, this.stateVectors.get(desc._sdr)));
            this.processYjsUpdate(napicli.util.b64ToUint8Array(desc._diff));
            //log(this.myPeerName+"("+this.ydocname+")", "rcv_reqdiff sdr=" + desc._sdr + ",size="+diff.length + ",total="+this.ydoc.getText('quill').length);
            this.stateVectors.delete(desc._sdr);
        }
        
        else if (desc._ytype === "repdiff") {
            const diff = napicli.util.b64ToUint8Array(desc._diff);
            this.versionVector.merge(new VersionVector(desc._vv)); // prepares synchro with relay
            this.processYjsUpdate(diff); // no impact on the version vector
            //log(this.myPeerName+"("+this.ydocname+")", "Updating my hash after synchro with replica: "+ this.stateHash);
            //log(this.myPeerName+"("+this.ydocname+")", "rcv_repdiff sdr=" + desc._sdr + ",size="+diff.length + ",total="+this.ydoc.getText('quill').length); 
        }
             
        /////////////////////////////////////// SYNCHRO WITH RELAY ///////////////////////////////////////////
        
        else if (desc._ytype === "strep" ) {
            if (desc._stpayld !== undefined) { 
                // Merge our version vector with the incoming one
                this.versionVector.merge(new VersionVector(desc._vv));
                //log(this.myPeerName+"("+this.ydocname+")", "Updating (merge) version vector to "+this.versionVector.toString());
                const upd = napicli.util.b64ToUint8Array(desc._stpayld);
                // Merge our state with the incoming one
                //log(this.myPeerName+"("+this.ydocname+")", "Updating state with "+desc._stpayld);
                this.processYjsUpdate(upd); // no impact on the version vector
            }
            if (desc._va !== undefined) { // we have processed all the states from the relay (if any), 
                const nbrel = Number.parseInt(desc._nbst);
                const va = new VersionVector(desc._va);
                // Check if we must send our state: 
                // not useful if the relay has in its store only one state and the store va is equal to the version vector
                // The version vector cannot be empty (we were not empty or we received at least a state, cf test before sendvv) 
                //console.log("Check if I must send my state.  vv==va:"+ this.versionVector.equals(va)+ "  nbrel:"+nbrel);
                if (! (this.versionVector.equals(va) && nbrel === 1)) {
                    const myState = Y.encodeStateAsUpdate(this.ydoc);
                    this.sendStateToOverwrite(desc._sdr, myState, this.versionVector);
                }
                
            }      
        }
        
        else log(this.myPeerName+"("+this.ydocname+")", "unknown received ytype message: "+desc._ytype);
        
        log(this.myPeerName+"("+this.ydocname+")", "My state after handling message: hash="+ this.stateHash+ "   vv="+this.versionVector.toString());
        
    }
     
         
    
    /*
     * Processes a document update received from another node
     * @param {Arraybuffer} update - the encoded update
     */
    processYjsUpdate(update) {
        //log(this.myPeerName+"("+this.ydocname+")", "applying update on Yjs doc ");
        Y.applyUpdate(this.ydoc, update, this);
    } 
    

    /**
     * Gives the version vector associated to the current state, 
     * formatted a string.
     * @returns {String}
     */
    getStringVersionVector() {
        return this.versionVector.toString();
    }


   /*
    * Destroys the provider 
    */
    destroyMyself() {
        log(this.myPeerName+"("+this.ydocname+")","destroyMyself");
        if ( ! this.shouldConnect) this.endNapiSession();
    }
    
    
   /*
    * Handlers for peer events
    */
    addPeer = (peer) => {
        log(this.myPeerName+"("+this.ydocname+")", "addPeer " + peer);
        if (isRelay(peer)) {
            this.synchroWithRelay(peer);
        }
        else {
          if (peer < this.myPeerName) this.exchangeDeltaWithPeer(peer);
        }
    }


    removePeer = (peer) => {
        clearTimeout(this.pendingSynchroPeers.get(peer));
        log(this.myPeerName+"("+this.ydocname+")", "removePeer " + peer);
    }


    clearPeers = () => { 
        this.pendingSynchroPeers.foreach((value) => clearTimeout(value));
    }

    
    

    /**
     * Sends a messge to a peer
     * @param {string} peer - The destinee peer
     * @param {Descriptor} desc - The message as a descriptor
     * @returns {undefined}
     */
    sendToPeer(peer, desc) {
        log("Sending to "+ peer + " : " +JSON.stringify(desc));
        let command = new napicli.napiConn.Pdu();
        command.dest = peer;
        command.name = "send_to_peer";
        command.tkn = napicli.util.uniqueId(); 
        command.desc = desc;
        this.session.sendCommand(command);
    }


/*
    broadcastToPeers(desc) {
        let command = new napicli.napiConn.Pdu();
        command.name = "broadcast_to_peers";
        command.tkn = napicli.util.uniqueId();
        command.desc = desc;
        this.session.sendCommand(command);
    }
 */   

    /**
     * Broadcasts a message to a selected subset of the known peers
     * @param {Descriptor} desc - The message as a descriptor
     * @param {function(string):boolean} filter - The function that selects a peer
     * @returns {undefined}
     */
    broadcastToPeers(desc, filter) {
        for (const peer of this.peerService.peerSet) {
            if (filter(peer)) {
                this.sendToPeer(peer, desc);
            } 
        }
    }
    
     
    
    
    /**
     * Initiates the synchro with a relay peer by sending to it my version vector
     * @param {string} peer - The relay peer 
     * @returns {undefined}
     */
    exchangeStatesWithRelay(peer) {
        if (this.shouldconnect) { log(this.myPeerName+"("+this.ydocname+")", "exchangeStatesWithRelay canceled (not connected)"); return;};
        //log(this.myPeerName+"("+this.ydocname+")", "exchangeStatesWithRelay " + peer);
        let desc = new napicli.napiPubsub.Descriptor();
        desc["_type"] = "yjs";
        desc["_endpoint"] = "yjsrelay";  
        desc["_ydocname"] = this.ydocname;  
        desc["_ytype"] = "sendvv";
        desc["_vv"] = this.versionVector.toString(); 
        this.sendToPeer(peer, desc); 
    }
    
    
    /**
     * Sends a state and its associated version vector to a relay peer
     * @param {string} peer - The relay peer
     * @param {uint8Array} st - the encoded state 
     * @param {string} vv  - The version vector
     * @returns {undefined}
     */
    sendStateToOverwrite(peer, st, vv) {
        //log(this.myPeerName+"("+this.ydocname+")", "sendStateToOverwrite to "+peer);
        let desc = new napicli.napiPubsub.Descriptor();
        desc["_type"] = "yjs";
        desc["_endpoint"] = "yjsrelay";  
        desc["_ydocname"] = this.ydocname;  
        desc["_ytype"] = "stowr";
        desc["_stpayld"] = napicli.util.uint8ArrayToB64(st);
        desc["_vv"] = vv.toString();
        this.sendToPeer(peer, desc); 
    }
    
    
    
}



function isNotRelay(peer) {
    return (! isRelay(peer));
}
